schedule = dict(
    type='string',
    regex=r'(\d+) (days|hours|minutes|seconds)',
)


schema = dict(
    connections=dict(
        type='dict',
        required=True,
        nullable=False,
        keyschema=dict(
            type='string',
            regex='[a-z\_]+',
        ),
    ),

    datasets=dict(
        type='dict',
        required=True,
        nullable=False,
        keyschema=dict(
            type='string',
            regex='[a-z\-]+',
        ),

        valueschema=dict(
            type='dict',
            schema=dict(
                name=dict(type='string'),

                query=dict(
                    type='string',
                    required=True,
                    nullable=False,
                ),

                graph=dict(
                    type='dict',
                    required=False,
                    nullable=True,
                ),

                schedule=schedule,
            ),
        ),
    ),

    dashboards=dict(
        type='dict',
        required=False,
        nullable=False,
        keyschema=dict(
            type='string',
            regex='[a-z\-]+',
        ),

        valueschema=dict(
            type='dict',
            schema=dict(
                name=dict(type='string'),

                datasets=dict(
                    type='list',
                    required=True,
                    nullable=False,
                ),
            ),
        ),
    ),

    schedule=schedule,
)



import click

from .app import (
    API_URL,
    App,
)


@click.group()
@click.option(
    '--config',
    '-c',
    required=False,
    default=".vizbee.yml",
    help="The YAML configuration file path.",
)
@click.option(
    '--client-id',
    envvar='CLIENT_ID',
    help="The application id.",
)
@click.option(
    '--client-secret',
    envvar='CLIENT_SECRET',
    help="The application secret.",
)
@click.option(
    '--api-url',
    envvar='API_URL',
    default=API_URL,
    help="The api url.",
)
@click.pass_context
def cli(context, config, client_id, client_secret, api_url):
    app = App(api_url, client_id, client_secret, context, config)
    context.obj = app


@cli.group('dataset')
def dataset():
    """Manage datasets."""
    pass


def _list(app, type_, remote=False):
    kpis = getattr(app, f"{type_}s").keys()

    if remote:
        kpis = app.list(type_)

    for key in kpis:
        click.echo(key)


def _prune(app, type_, force=False):
    orphans = app.orphans(type_)

    if len(orphans) == 0:
        app.log("Remote is up to date", level="info")
        return True

    slugs = ", ".join(orphans)

    if not force and not click.confirm(
        click.style(f"You're about to delete: {slugs}\n", fg='red')
        + 'Do you want to continue?'
    ):
        return False

    for orphan in orphans:
        app.delete(type_, orphan)


def _push(app, type_, slug, open):
    item = app.get(type_, slug)
    return item.push(open)


@dataset.command()
@click.argument('dataset')
@click.option('--json', is_flag=True)
@click.pass_obj
def execute(app, dataset, json):
    """Execute given dataset query."""
    dataset = app.get('dataset', dataset)
    results = dataset.execute()
    dataset = results.dataset

    if json:
        dataset = dataset.json

    click.echo(dataset)


@dataset.command(name='list')
@click.option('--remote', is_flag=True)
@click.pass_obj
def dataset_list(app, remote=False):
    """List available datasets."""
    return _list(app, 'dataset', remote=remote)


@dataset.command(name='push')
@click.argument('dataset')
@click.option('--open', is_flag=True)
@click.pass_obj
def dataset_push(app, dataset, open):
    """Push given dataset."""
    return _push(app, 'dataset', dataset, open)


@dataset.command(name='prune')
@click.option('--force', is_flag=True)
@click.pass_obj
def dataset_prune(app, force):
    """Delete remote orphan datasets."""
    _prune(app, 'dataset', force=force)


@dataset.command()
@click.argument('dataset')
@click.option('--execute', is_flag=True)
@click.pass_obj
def dataset_edit(app, dataset, execute):
    """Edit given dataset query."""
    # TODO: save
    dataset = app.get('dataset', dataset)
    dataset.query = click.edit(dataset.query, extension='.sql')
    results = dataset.execute()
    click.echo(results.dataset)


@cli.group('dashboard')
def dashboard():
    """Manage dashboards."""
    pass


@dashboard.command(name='list')
@click.option('--remote', is_flag=True)
@click.pass_obj
def dashboard_list(app, remote=False):
    """List available dashboards."""
    return _list(app, 'dashboard', remote=remote)


@dashboard.command(name='push')
@click.argument('dashboard')
@click.option('--open', is_flag=True)
@click.pass_obj
def dashboard_push(app, dashboard, open):
    """Push given dashboard."""
    return _push(app, 'dashboard', dashboard, open)


@dashboard.command(name='prune')
@click.option('--force', is_flag=True)
@click.pass_obj
def dashboard_prune(app, force):
    """Delete remote orphan dashboards."""
    _prune(app, 'dashboard', force=force)


@cli.command()
@click.pass_obj
def sync(app):
    """Push all datasets and dashboards."""
    return app.sync()


@cli.command()
@click.pass_obj
def start(app):
    """Start scheduler."""
    app.start()


if __name__ == '__main__':
    cli.main()



from setuptools import setup
from vizbee import __version__


with open('README.md', 'r') as f:
    setup(
        name='vizbee',
        version=__version__,
        description='vizbee.io agent',
        long_description=f.read(),
        author='jean-philippe serafin',
        author_email='serafinjp@gmail.com',
        url='https://gitlab.com/vizbee/agent',
        packages=['vizbee'],
        install_requires=[
            'apscheduler',
            'cerberus',
            'click',
            'pyaml',
            'records',
            'requests',
            'simplejson',
        ],
        license='MIT',
        entry_points=dict(
            console_scripts=['vizbee=vizbee.cli:cli.main'],
        ),
        tests_require=['responses'],
        test_suite="vizbee.tests.suite",
        extras_require=dict(
            postgresql=['psycopg2'],
            oracle=['cx_oracle'],
            mysql=['mysqlclient'],
            redshift=['sqlalchemy-redshift'],
            sqlserver=['pyodbc'],
            sybase=['pyodbc']
        ),
    )



import os
import logging

from collections import OrderedDict

import click
import records
import requests

from apscheduler.schedulers.blocking import BlockingScheduler
from cerberus import Validator
from sqlalchemy.exc import DatabaseError

from yaml import load, dump
from yaml.error import YAMLError

from .schema import schema


logger = logging.getLogger(__name__)


API_URL = 'https://api.vizbee.io/v1'


class Item():
    @property
    def url_prefix(self):
        raise NotImplementedError()

    def log(self, message, **kwargs):
        slug = click.style(f"`{self.slug}`", fg='cyan')
        kwargs.update(dict(slug=slug))
        self.app.log(message, **kwargs)

    def push(self, open_=False):
        payload = self.payload

        self.log("Pushing: {slug}")

        response = self.app.request(
            f'/{self.url_prefix}/{self.slug}',
            data=payload,
        )

        status = response.status_code

        if status not in (200, 201):
            if status == 422:
                json = response.json()
                errors = json['errors']

                errors = self.app.format_errors(errors)

                self.log(
                    "Errors: {errors}",
                    level='warning',
                    errors=click.style(str(errors), fg='red')
                )
                return False

            self.log(
                "The server unexpectedly responded with `{status}` status",
                status=str(status),
                level='warning',
            )
            return False

        json = response.json()
        verb = 'created' if status == 201 else 'updated'
        url = json['url']
        self.log(
            "Successfully {verb} {slug}: {url}",
            verb=verb,
            url=click.style(url, fg='white')
        )

        if open_:
            click.launch(url)

        return True


class Dataset(Item):
    url_prefix = "datasets"

    def __init__(
        self,
        app,
        slug,
        query,
        connection,
        graph,
        name=None,
        schedule=None,
    ):
        self.app = app
        self.slug = slug
        self.query = query
        self.connection = connection
        self.graph = graph
        self.name = name

        if schedule is None:
            schedule = app.schedule

        self.schedule = schedule

    @property
    def payload(self):
        data = self.execute().dataset.dict

        return dict(
            name=self.name,
            graph=self.graph,
            query=self.query,
            data=data,
        )

    def execute(self):
        self.log("Executing: {slug}")
        try:
            return self.connection.query(self.query)

        except DatabaseError as e:
            self.log(str(e), level='critical')

    def schedule_job(self, scheduler):
        schedule = self.schedule

        if schedule is None:
            raise ValueError(
                f"No scheduling rule found for `{self.slug}`"
            )

        duration, unit = schedule.split(' ')

        scheduler.add_job(self.push, 'interval', **{unit: int(duration)})


class Dashboard(Item):
    url_prefix = "dashboards"

    def __init__(self, app, slug, datasets, name=None):
        self.app = app
        self.slug = slug
        self.name = name
        self.datasets = datasets

    @property
    def payload(self):
        return dict(
            name=self.name,
            datasets=self.datasets,
        )


class App():
    def __init__(self, api_url, client_id, client_secret, cli, filename):
        self.api_url = api_url
        self.client_id = client_id
        self.client_secret = client_secret
        self.cli = cli
        self.daemonized = False

        config = self.load_config(filename)
        connections = config['connections']

        try:
            self.connections = {
                key: records.Database(url)
                for key, url in connections.items()
            }

        except Exception as e:
            self.log(
                "Error initializing connections: {e}",
                e=e,
                level='critical',
            )

        self.schedule = config.get('schedule')

        datasets = OrderedDict()

        for key, dataset in config['datasets'].items():
            datasets[key] = Dataset(
                self,
                key,
                dataset['query'],
                self.connections[dataset.get('connection', 'default')],
                dataset.get('graph'),
                dataset.get('name'),
                dataset.get('schedule'),
            )

        self.datasets = datasets

        dashboards = OrderedDict()

        configured_dashboards = config.get('dashboards')

        if configured_dashboards is not None:
            for key, dashboard in config['dashboards'].items():
                dashboards[key] = Dashboard(
                    self,
                    key,
                    dashboard.get('datasets'),
                    dashboard.get('name'),
                )

        self.dashboards = dashboards

    def format_errors(self, errors):
        if isinstance(errors, dict):
            errors = dump(errors)

        errors = str(errors)

        if self.daemonized:
            return errors

        return f"\n\n{errors}"

    def request(self, url, method='put', data=None):
        headers = {
            'Content-type': 'application/json',
            'Accept': 'application/json',
        }

        from requests.exceptions import RequestException

        try:
            return getattr(requests, method)(
                self.api_url + url,
                json=data,
                auth=(self.client_id, self.client_secret),
                headers=headers,
                allow_redirects=False,
            )

        except RequestException as e:
            self.log(
                "Error sending request: \n{e}",
                e=str(e),
                level='critical',
            )

    def load_config(self, filename):
        try:
            with open(filename, 'r') as f:
                template = "".join(f.readlines())
                template = template.format(**os.environ)
                config = load(template)

                validator = Validator()

                if not validator.validate(config, schema):
                    raise ValueError(validator.errors)

                self.schema = schema
                return config

        except FileNotFoundError:
            self.log(
                "`{filename}` not found",
                filename=filename,
                level='critical',
            )

        except (YAMLError, ValueError) as e:
            self.log(
                "Error parsing config: {e}",
                e=self.format_errors(e),
                level='critical',
            )

        except KeyError as e:
            self.log(
                "Error parsing config: \n{e} not found in environment.",
                e=str(e),
                level='critical',
            )

    def log(self, message, level='info', **context):
        message = message.format(**context)

        if level in ('critical', 'warning'):
            message = click.style(message, fg='red')

        if self.daemonized:
            getattr(logger, level)(message)

        else:
            click.echo(message)

        if level == 'critical':
            self.cli.exit(1)

    def get(self, type_, slug):
        try:
            return getattr(self, f"{type_}s")[slug]

        except KeyError:
            self.log(
                "Invalid {type_} `{slug}`",
                slug=slug,
                type_=type_,
                level='critical',
            )

    def list(self, type_):
        response = self.request(f'/{type_}s/', method='get')
        json = response.json()
        return [
           item['slug'] for item in json
        ]

    def orphans(self, type_):
        remotes = self.list(type_)
        return sorted(
            set(remotes) - set(
                getattr(self, f"{type_}s").keys()
            )
        )

    def delete(self, type_, slug):
        self.log(
            "Deleting {type_} `{slug}`",
            slug=slug,
            type_=type_,
            level='info',
        )
        self.request(f'/{type_}s/{slug}', method='delete')

    def start(self, sync=True):
        scheduler = BlockingScheduler()

        try:
            for dataset in self.datasets.values():
                dataset.schedule_job(scheduler)

        except ValueError as e:
            self.log(
                str(e),
                level='critical',
            )

        if sync:
            self.log("Triggering initial sync", level='info')
            if not self.sync():
                return False

        self.log("Start processing jobs", level='info')
        scheduler.start()

    def sync(self):
        for collection in (self.datasets, self.dashboards):
            for item in collection.values():
                if not item.push():
                    self.log("Sync failed", level='critical')
                    return False

        return True
